"""
Module to provide tests related to parsing from stdin.
"""

import os
from test.markdown_scanner import MarkdownScanner


def test_markdown_with_scan_stdin_without_triggers():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan-stdin",
    ]

    supplied_standard_input = "test\nme\n"
    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, standard_input_to_use=supplied_standard_input
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_scan_stdin_with_triggers():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan-stdin",
    ]

    supplied_standard_input = "# test"
    expected_return_code = 1
    expected_output = """stdin:1:1: MD022: Headings should be surrounded by blank lines. [Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)
stdin:1:6: MD047: Each file should end with a single newline character. (single-trailing-newline)"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, standard_input_to_use=supplied_standard_input
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_scan_stdin_with_triggers_and_disabled_rules():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "-d",
        "MD022, MD047",
        "scan-stdin",
    ]

    supplied_standard_input = "# test"
    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, standard_input_to_use=supplied_standard_input
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_scan_stdin_with_bad_write():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "-x-stdin",
        "scan-stdin",
    ]

    supplied_standard_input = f"# test{os.linesep}"
    expected_return_code = 1
    expected_output = ""
    expected_error = """OSError encountered while scanning 'stdin':
Temporary file to capture stdin was not written (made up)."""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, standard_input_to_use=supplied_standard_input
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
Module to provide tests related to alternate extensions.
"""

import os
from test.markdown_scanner import MarkdownScanner


def test_markdown_with_dash_ae_with_invalid_file_extension():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    file_to_scan = os.path.join(
        "test", "resources", "double-line-with-blank-and-trailing.txt"
    )
    supplied_arguments = [
        "scan",
        "-ae",
        ".abc",
        file_to_scan,
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = f"Provided file path '{file_to_scan}' is not a valid file. Skipping.\n\n\nNo matching files found."

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_ae_with_valid_file_extension():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    file_to_scan = os.path.join(
        "test", "resources", "double-line-with-blank-and-trailing.txt"
    )
    supplied_arguments = [
        "scan",
        "-ae",
        ".txt",
        file_to_scan,
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_ae_with_valid_file_extension_multiple():
    """
    Test to make sure that a file with a non-md extension but with
    alternate extensions is accepted as a valid file.

    This function is a shadow of
    test_api_list_for_non_markdown_file_with_alternate_extensions
    """

    # Arrange
    scanner = MarkdownScanner()
    file_to_scan = "test/resources/double-line-with-blank-and-trailing.txt"
    supplied_arguments = [
        "scan",
        "-ae",
        ".txt,.md",
        file_to_scan,
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_ae_with_invalid_file_extension_no_period():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    file_to_scan = os.path.join(
        "test", "resources", "double-line-with-blank-and-trailing.txt"
    )
    supplied_arguments = [
        "scan",
        "-ae",
        "md",
        file_to_scan,
    ]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py scan [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS]
                    [-e PATH_EXCLUSIONS]
                    path [path ...]
main.py scan: error: argument -ae/--alternate-extensions: Extension 'md' must start with a period."""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_ae_with_invalid_file_extension_no_alphanum():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    file_to_scan = os.path.join(
        "test", "resources", "double-line-with-blank-and-trailing.txt"
    )
    supplied_arguments = [
        "scan",
        "-ae",
        ".*",
        file_to_scan,
    ]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py scan [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS]
                    [-e PATH_EXCLUSIONS]
                    path [path ...]
main.py scan: error: argument -ae/--alternate-extensions: Extension '.*' must only contain alphanumeric characters after the period."""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_ae_with_invalid_file_extension_only_period():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    file_to_scan = os.path.join(
        "test", "resources", "double-line-with-blank-and-trailing.txt"
    )
    supplied_arguments = [
        "scan",
        "-ae",
        ".",
        file_to_scan,
    ]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py scan [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS]
                    [-e PATH_EXCLUSIONS]
                    path [path ...]
main.py scan: error: argument -ae/--alternate-extensions: Extension '.' must have at least one character after the period."""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_ae_with_invalid_file_extension_semicolon_as_sep():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    file_to_scan = os.path.join(
        "test", "resources", "double-line-with-blank-and-trailing.txt"
    )
    supplied_arguments = [
        "scan",
        "-ae",
        ".md;.txt",
        file_to_scan,
    ]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py scan [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS]
                    [-e PATH_EXCLUSIONS]
                    path [path ...]
main.py scan: error: argument -ae/--alternate-extensions: Extension '.md;.txt' must only contain alphanumeric characters after the period."""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_ae_with_invalid_file_extension_empty():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    file_to_scan = os.path.join(
        "test", "resources", "double-line-with-blank-and-trailing.txt"
    )
    supplied_arguments = [
        "scan",
        "-ae",
        "",
        file_to_scan,
    ]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py scan [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS]
                    [-e PATH_EXCLUSIONS]
                    path [path ...]
main.py scan: error: argument -ae/--alternate-extensions: Extension '' must start with a period."""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_ae_xxx1():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    directory_to_scan = os.path.join("test", "resources", "test-directory-1")
    supplied_arguments = [
        "scan",
        "-l",
        "--alternate-extension=.qmd",
        f"{directory_to_scan}/*",
    ]

    expected_return_code = 0
    expected_output = f"{directory_to_scan}{os.sep}test.qmd"
    expected_error = f"Provided file path '{directory_to_scan}{os.sep}README.md' is not a valid file. Skipping."

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_ae_xxx2():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    directory_to_scan = os.path.join("test", "resources", "test-directory-1")
    supplied_arguments = [
        "scan",
        "-l",
        "--alternate-extension=.qmd",
        f"{directory_to_scan}/",
    ]

    expected_return_code = 0
    expected_output = f"{directory_to_scan}{os.sep}test.qmd"
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
Module to provide tests related to enabling or disabling rules.
"""

import os
from test.markdown_scanner import MarkdownScanner
from test.utils import create_temporary_configuration_file


def test_markdown_with_dash_e_single_by_name():
    """
    Test to make sure we get enable a rule if '-e' is supplied and the name of the
    rule is provided. The test data for MD047 is used as it is a simple file that
    passes normally, it is used as a comparison.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "-e",
        "debug-only",
        "scan",
        source_path,
    ]

    expected_return_code = 0
    expected_output = """MD999>>init_from_config
MD999>>test_value>>1
MD999>>other_test_value>>1
MD999>>starting_new_file>>
MD999>>token:[atx(1,1):1:0:]
MD999>>token:[text(1,3):This is a test: ]
MD999>>token:[end-atx::]
MD999>>token:[BLANK(2,1):]
MD999>>token:[para(3,1):]
MD999>>token:[text(3,1):The line after this line should be blank.:]
MD999>>token:[end-para:::True]
MD999>>token:[BLANK(4,1):]
MD999>>token:[end-of-stream(5,0)]
MD999>>next_line:# This is a test
MD999>>next_line:
MD999>>next_line:The line after this line should be blank.
MD999>>next_line:
MD999>>completed_file
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_e_single_by_id():
    """
    Test to make sure we get enable a rule if '-e' is supplied and the id of the
    rule is provided. The test data for MD047 is used as it is a simple file that
    passes normally, it is used as a comparison.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "-e",
        "MD999",
        "scan",
        source_path,
    ]

    expected_return_code = 0
    expected_output = """MD999>>init_from_config
MD999>>test_value>>1
MD999>>other_test_value>>1
MD999>>starting_new_file>>
MD999>>token:[atx(1,1):1:0:]
MD999>>token:[text(1,3):This is a test: ]
MD999>>token:[end-atx::]
MD999>>token:[BLANK(2,1):]
MD999>>token:[para(3,1):]
MD999>>token:[text(3,1):The line after this line should be blank.:]
MD999>>token:[end-para:::True]
MD999>>token:[BLANK(4,1):]
MD999>>token:[end-of-stream(5,0)]
MD999>>next_line:# This is a test
MD999>>next_line:
MD999>>next_line:The line after this line should be blank.
MD999>>next_line:
MD999>>completed_file
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_enabled_by_configuration_id():
    """
    Test to make sure we enable a rule by using the rule's id in the
    configuration, with no help from the command line.
    The test data for MD047 is used as it is a simple file that
    passes normally, it is used as a comparison.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"plugins": {"md999": {"enabled": True}}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 0
        expected_output = """MD999>>init_from_config
MD999>>test_value>>1
MD999>>other_test_value>>1
MD999>>starting_new_file>>
MD999>>token:[atx(1,1):1:0:]
MD999>>token:[text(1,3):This is a test: ]
MD999>>token:[end-atx::]
MD999>>token:[BLANK(2,1):]
MD999>>token:[para(3,1):]
MD999>>token:[text(3,1):The line after this line should be blank.:]
MD999>>token:[end-para:::True]
MD999>>token:[BLANK(4,1):]
MD999>>token:[end-of-stream(5,0)]
MD999>>next_line:# This is a test
MD999>>next_line:
MD999>>next_line:The line after this line should be blank.
MD999>>next_line:
MD999>>completed_file
"""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_enabled_by_configuration_name():
    """
    Test to make sure we enable a rule by using the rule's name in the
    configuration, with no help from the command line.
    The test data for MD047 is used as it is a simple file that
    passes normally, it is used as a comparison.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"plugins": {"debug-only": {"enabled": True}}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-c",
            configuration_file,
            # "--log-level",
            # "DEBUG",
            "scan",
            source_path,
        ]

        expected_return_code = 0
        expected_output = """MD999>>init_from_config
MD999>>test_value>>1
MD999>>other_test_value>>1
MD999>>starting_new_file>>
MD999>>token:[atx(1,1):1:0:]
MD999>>token:[text(1,3):This is a test: ]
MD999>>token:[end-atx::]
MD999>>token:[BLANK(2,1):]
MD999>>token:[para(3,1):]
MD999>>token:[text(3,1):The line after this line should be blank.:]
MD999>>token:[end-para:::True]
MD999>>token:[BLANK(4,1):]
MD999>>token:[end-of-stream(5,0)]
MD999>>next_line:# This is a test
MD999>>next_line:
MD999>>next_line:The line after this line should be blank.
MD999>>next_line:
MD999>>completed_file
"""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_dash_d_single_by_name():
    """
    Test to make sure we get enable a rule if '-d' is supplied and the name of the
    rule is provided. The test data for MD047 is used as it is a simple file that
    fails normally, it is used as a comparison.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "-d",
        "single-trailing-newline",
        "scan",
        source_path,
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_d_single_by_id():
    """
    Test to make sure we get enable a rule if '-d' is supplied and the id of the
    rule is provided. The test data for MD047 is used as it is a simple file that
    fails normally, it is used as a comparison.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "-d",
        "MD047",
        "scan",
        source_path,
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_d_and_dash_e_single_by_name():
    """
    Test to make sure we get disabled if a rule if '-d' is supplied
    and if 'e' is supplied, both with the name of the rule.
    The test data for MD047 is used as it is a simple file that
    fails normally, it is used as a comparison.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "-d",
        "single-trailing-newline",
        "-e",
        "single-trailing-newline",
        "scan",
        source_path,
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
Module to provide tests for altering how the command line handles exceptions.
"""

import os
from test.markdown_scanner import MarkdownScanner
from test.utils import create_temporary_configuration_file


def test_exception_handling_no_exception():
    """
    Test to validate normal operations of scanning files without any plugin exception being thrown.
    """

    contents_file_1_and_3 = """#\tPerfectly healthy file
This triggers several rules:
1. Guess which ones
1. Bla
"""

    with create_temporary_configuration_file(
        contents_file_1_and_3, file_name_prefix="tmp1", file_name_suffix=".md"
    ) as file_name_1:
        with create_temporary_configuration_file(
            contents_file_1_and_3, file_name_prefix="tmp3", file_name_suffix=".md"
        ) as file_name_3:
            # Arrange
            scanner = MarkdownScanner()
            supplied_arguments = ["scan", file_name_1, file_name_3]

            expected_return_code = 1
            expected_output = """{file_name_1}:1:1: MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)
{file_name_1}:1:1: MD022: Headings should be surrounded by blank lines. [Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)
{file_name_1}:1:2: MD010: Hard tabs [Column: 2] (no-hard-tabs)
{file_name_1}:3:1: MD032: Lists should be surrounded by blank lines (blanks-around-lists)
{file_name_3}:1:1: MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)
{file_name_3}:1:1: MD022: Headings should be surrounded by blank lines. [Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)
{file_name_3}:1:2: MD010: Hard tabs [Column: 2] (no-hard-tabs)
{file_name_3}:3:1: MD032: Lists should be surrounded by blank lines (blanks-around-lists)
""".replace(
                "{file_name_1}", file_name_1
            ).replace(
                "{file_name_3}", file_name_3
            )
            expected_error = ""

            # Act
            execute_results = scanner.invoke_main(arguments=supplied_arguments)

            # Assert
            execute_results.assert_results(
                expected_output, expected_error, expected_return_code
            )


def test_exception_handling_scan_with_plugin_exception_and_no_flag():
    """
    Test to validate normal operations of scanning files with a simple plugin exception being thrown.
    """

    contents_file_1_and_3 = """#\tPerfectly healthy file
This triggers several rules:
1. Guess which ones
1. Bla
"""
    contents_file_2 = """# File the causes exception

throw_exception
"""
    plugin_path = os.path.join(
        "test", "resources", "plugins", "bad", "bad_next_line_with_scan_trigger.py"
    )

    with create_temporary_configuration_file(
        contents_file_1_and_3, file_name_prefix="tmp1", file_name_suffix=".md"
    ) as file_name_1:
        with create_temporary_configuration_file(
            contents_file_2, file_name_prefix="tmp2", file_name_suffix=".md"
        ) as file_name_2:
            with create_temporary_configuration_file(
                contents_file_1_and_3, file_name_prefix="tmp3", file_name_suffix=".md"
            ) as file_name_3:
                # Arrange
                scanner = MarkdownScanner()
                supplied_arguments = [
                    "--add-plugin",
                    plugin_path,
                    "scan",
                    file_name_1,
                    file_name_2,
                    file_name_3,
                ]

                expected_return_code = 1
                expected_output = """{file_name_1}:1:1: MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)
{file_name_1}:1:1: MD022: Headings should be surrounded by blank lines. [Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)
{file_name_1}:1:2: MD010: Hard tabs [Column: 2] (no-hard-tabs)
{file_name_1}:3:1: MD032: Lists should be surrounded by blank lines (blanks-around-lists)
""".replace(
                    "{file_name_1}", file_name_1
                ).replace(
                    "{file_name_3}", file_name_3
                )
                expected_error = """
                
BadPluginError encountered while scanning '{file_name_2}':
(Line 3): Plugin id 'MDE008' had a critical failure during the 'next_line' action.
""".replace(
                    "{file_name_2}", file_name_2
                )
                # Act
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

                # Assert
                execute_results.assert_results(
                    expected_output, expected_error, expected_return_code
                )


def test_exception_handling_scan_with_plugin_exception_and_flag():
    """
    Test to validate normal operations of scanning files with a simple plugin exception being thrown.
    """

    contents_file_1_and_3 = """#\tPerfectly healthy file
This triggers several rules:
1. Guess which ones
1. Bla
"""
    contents_file_2 = """# File the causes exception

throw_exception
"""
    plugin_path = os.path.join(
        "test", "resources", "plugins", "bad", "bad_next_line_with_scan_trigger.py"
    )

    with create_temporary_configuration_file(
        contents_file_1_and_3, file_name_prefix="tmp1", file_name_suffix=".md"
    ) as file_name_1:
        with create_temporary_configuration_file(
            contents_file_2, file_name_prefix="tmp2", file_name_suffix=".md"
        ) as file_name_2:
            with create_temporary_configuration_file(
                contents_file_1_and_3, file_name_prefix="tmp3", file_name_suffix=".md"
            ) as file_name_3:
                # Arrange
                scanner = MarkdownScanner()
                supplied_arguments = [
                    "--add-plugin",
                    plugin_path,
                    "--continue-on-error",
                    "scan",
                    file_name_1,
                    file_name_2,
                    file_name_3,
                ]

                expected_return_code = 1
                expected_output = """{file_name_1}:1:1: MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)
{file_name_1}:1:1: MD022: Headings should be surrounded by blank lines. [Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)
{file_name_1}:1:2: MD010: Hard tabs [Column: 2] (no-hard-tabs)
{file_name_1}:3:1: MD032: Lists should be surrounded by blank lines (blanks-around-lists)
{file_name_3}:1:1: MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)
{file_name_3}:1:1: MD022: Headings should be surrounded by blank lines. [Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)
{file_name_3}:1:2: MD010: Hard tabs [Column: 2] (no-hard-tabs)
{file_name_3}:3:1: MD032: Lists should be surrounded by blank lines (blanks-around-lists)
""".replace(
                    "{file_name_1}", file_name_1
                ).replace(
                    "{file_name_3}", file_name_3
                )
                expected_error = """{file_name_2}:0:0: (Line 3): Plugin id 'MDE008' had a critical failure during the 'next_line' action.
""".replace(
                    "{file_name_2}", file_name_2
                )
                # Act
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

                # Assert
                execute_results.assert_results(
                    expected_output, expected_error, expected_return_code
                )


def test_exception_handling_scan_with_tokenization_exception_and_no_flag():
    """
    Test to validate normal operations of scanning files with a simple tokenization exception being thrown.
    """

    contents_file_1_and_3 = """#\tPerfectly healthy file
This triggers several rules:
1. Guess which ones
1. Bla
"""
    extension_enable_front_matter = "extensions.front-matter.enabled=$!True"
    contents_file_2 = """---
test: assert
---
"""
    with create_temporary_configuration_file(
        contents_file_1_and_3, file_name_prefix="tmp1", file_name_suffix=".md"
    ) as file_name_1:
        with create_temporary_configuration_file(
            contents_file_2, file_name_prefix="tmp2", file_name_suffix=".md"
        ) as file_name_2:
            with create_temporary_configuration_file(
                contents_file_1_and_3, file_name_prefix="tmp3", file_name_suffix=".md"
            ) as file_name_3:
                # Arrange
                scanner = MarkdownScanner()
                supplied_arguments = [
                    "--set",
                    extension_enable_front_matter,
                    "scan",
                    file_name_1,
                    file_name_2,
                    file_name_3,
                ]

                expected_return_code = 1
                expected_output = """{file_name_1}:1:1: MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)
{file_name_1}:1:1: MD022: Headings should be surrounded by blank lines. [Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)
{file_name_1}:1:2: MD010: Hard tabs [Column: 2] (no-hard-tabs)
{file_name_1}:3:1: MD032: Lists should be surrounded by blank lines (blanks-around-lists)
""".replace(
                    "{file_name_1}", file_name_1
                ).replace(
                    "{file_name_3}", file_name_3
                )
                expected_error = """
                
Unexpected Error(BadTokenizationError): An unhandled error occurred processing the document.
""".replace(
                    "{file_name_2}", file_name_2
                )
                # Act
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

                # Assert
                execute_results.assert_results(
                    expected_output, expected_error, expected_return_code
                )


def test_exception_handling_scan_with_tokenization_exception_and_flag():
    """
    Test to validate normal operations of scanning files with a simple tokenization exception being thrown.
    """

    contents_file_1_and_3 = """#\tPerfectly healthy file
This triggers several rules:
1. Guess which ones
1. Bla
"""
    extension_enable_front_matter = "extensions.front-matter.enabled=$!True"
    contents_file_2 = """---
test: assert
---
"""
    with create_temporary_configuration_file(
        contents_file_1_and_3, file_name_prefix="tmp1", file_name_suffix=".md"
    ) as file_name_1:
        with create_temporary_configuration_file(
            contents_file_2, file_name_prefix="tmp2", file_name_suffix=".md"
        ) as file_name_2:
            with create_temporary_configuration_file(
                contents_file_1_and_3, file_name_prefix="tmp3", file_name_suffix=".md"
            ) as file_name_3:
                # Arrange
                scanner = MarkdownScanner()
                supplied_arguments = [
                    "--continue-on-error",
                    "--set",
                    extension_enable_front_matter,
                    "scan",
                    file_name_1,
                    file_name_2,
                    file_name_3,
                ]

                expected_return_code = 1
                expected_output = """{file_name_1}:1:1: MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)
{file_name_1}:1:1: MD022: Headings should be surrounded by blank lines. [Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)
{file_name_1}:1:2: MD010: Hard tabs [Column: 2] (no-hard-tabs)
{file_name_1}:3:1: MD032: Lists should be surrounded by blank lines (blanks-around-lists)
{file_name_3}:1:1: MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)
{file_name_3}:1:1: MD022: Headings should be surrounded by blank lines. [Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)
{file_name_3}:1:2: MD010: Hard tabs [Column: 2] (no-hard-tabs)
{file_name_3}:3:1: MD032: Lists should be surrounded by blank lines (blanks-around-lists)
""".replace(
                    "{file_name_1}", file_name_1
                ).replace(
                    "{file_name_3}", file_name_3
                )
                expected_error = """{file_name_2}:0:0: An unhandled error occurred processing the document.
""".replace(
                    "{file_name_2}", file_name_2
                )
                # Act
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

                # Assert
                execute_results.assert_results(
                    expected_output, expected_error, expected_return_code
                )


def test_exception_handling_fix_with_plugin_exception_and_no_flag():
    """
    Test to validate normal operations of scanning files with a simple plugin exception being thrown.
    """

    contents_file_1_and_3 = """#\tPerfectly healthy file
This triggers several rules:
1. Guess which ones
1. Bla
"""
    contents_file_2 = """# File the causes exception

throw_exception
"""
    plugin_path = os.path.join(
        "test", "resources", "plugins", "bad", "bad_next_line_with_fix_trigger.py"
    )

    with create_temporary_configuration_file(
        contents_file_1_and_3, file_name_prefix="tmp1", file_name_suffix=".md"
    ) as file_name_1:
        with create_temporary_configuration_file(
            contents_file_2, file_name_prefix="tmp2", file_name_suffix=".md"
        ) as file_name_2:
            with create_temporary_configuration_file(
                contents_file_1_and_3, file_name_prefix="tmp3", file_name_suffix=".md"
            ) as file_name_3:
                # Arrange
                scanner = MarkdownScanner()
                supplied_arguments = [
                    "--add-plugin",
                    plugin_path,
                    "fix",
                    file_name_1,
                    file_name_2,
                    file_name_3,
                ]

                expected_return_code = 1
                expected_output = """Fixed: {file_name_1}
""".replace(
                    "{file_name_1}", file_name_1
                ).replace(
                    "{file_name_3}", file_name_3
                )
                expected_error = """
                
BadPluginError encountered while scanning '{file_name_2}':
(Line 3): Plugin id 'MDE008' had a critical failure during the 'next_line' action.
""".replace(
                    "{file_name_2}", file_name_2
                )
                # Act
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

                # Assert
                execute_results.assert_results(
                    expected_output, expected_error, expected_return_code
                )


def test_exception_handling_fix_with_plugin_exception_and_flag():
    """
    Test to validate normal operations of scanning files with a simple plugin exception being thrown.
    """

    contents_file_1_and_3 = """#\tPerfectly healthy file
This triggers several rules:
1. Guess which ones
1. Bla
"""
    contents_file_2 = """# File the causes exception

throw_exception
"""
    plugin_path = os.path.join(
        "test", "resources", "plugins", "bad", "bad_next_line_with_fix_trigger.py"
    )

    with create_temporary_configuration_file(
        contents_file_1_and_3, file_name_prefix="tmp1", file_name_suffix=".md"
    ) as file_name_1:
        with create_temporary_configuration_file(
            contents_file_2, file_name_prefix="tmp2", file_name_suffix=".md"
        ) as file_name_2:
            with create_temporary_configuration_file(
                contents_file_1_and_3, file_name_prefix="tmp3", file_name_suffix=".md"
            ) as file_name_3:
                # Arrange
                scanner = MarkdownScanner()
                supplied_arguments = [
                    "--continue-on-error",
                    "--add-plugin",
                    plugin_path,
                    "fix",
                    file_name_1,
                    file_name_2,
                    file_name_3,
                ]

                expected_return_code = 1
                expected_output = """Fixed: {file_name_1}
Fixed: {file_name_3}
""".replace(
                    "{file_name_1}", file_name_1
                ).replace(
                    "{file_name_3}", file_name_3
                )
                expected_error = """{file_name_2}:0:0: (Line 3): Plugin id 'MDE008' had a critical failure during the 'next_line' action.
""".replace(
                    "{file_name_2}", file_name_2
                )
                # Act
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

                # Assert
                execute_results.assert_results(
                    expected_output, expected_error, expected_return_code
                )


def test_exception_handling_fix_with_tokenization_exception_and_no_flag():
    """
    Test to validate normal operations of scanning files with a simple tokenization exception being thrown.
    """

    contents_file_1_and_3 = """#\tPerfectly healthy file
This triggers several rules:
1. Guess which ones
1. Bla
"""
    extension_enable_front_matter = "extensions.front-matter.enabled=$!True"
    contents_file_2 = """---
test: assert
---
"""
    with create_temporary_configuration_file(
        contents_file_1_and_3, file_name_prefix="tmp1", file_name_suffix=".md"
    ) as file_name_1:
        with create_temporary_configuration_file(
            contents_file_2, file_name_prefix="tmp2", file_name_suffix=".md"
        ) as file_name_2:
            with create_temporary_configuration_file(
                contents_file_1_and_3, file_name_prefix="tmp3", file_name_suffix=".md"
            ) as file_name_3:
                # Arrange
                scanner = MarkdownScanner()
                supplied_arguments = [
                    "--set",
                    extension_enable_front_matter,
                    "fix",
                    file_name_1,
                    file_name_2,
                    file_name_3,
                ]

                expected_return_code = 1
                expected_output = """Fixed: {file_name_1}
""".replace(
                    "{file_name_1}", file_name_1
                ).replace(
                    "{file_name_3}", file_name_3
                )
                expected_error = """
                
Unexpected Error(BadTokenizationError): An unhandled error occurred processing the document.
""".replace(
                    "{file_name_2}", file_name_2
                )
                # Act
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

                # Assert
                execute_results.assert_results(
                    expected_output, expected_error, expected_return_code
                )


def test_exception_handling_fix_with_tokenization_exception_and_flag():
    """
    Test to validate normal operations of scanning files with a simple tokenization exception being thrown.
    """

    contents_file_1_and_3 = """#\tPerfectly healthy file
This triggers several rules:
1. Guess which ones
1. Bla
"""
    extension_enable_front_matter = "extensions.front-matter.enabled=$!True"
    contents_file_2 = """---
test: assert
---
"""
    with create_temporary_configuration_file(
        contents_file_1_and_3, file_name_prefix="tmp1", file_name_suffix=".md"
    ) as file_name_1:
        with create_temporary_configuration_file(
            contents_file_2, file_name_prefix="tmp2", file_name_suffix=".md"
        ) as file_name_2:
            with create_temporary_configuration_file(
                contents_file_1_and_3, file_name_prefix="tmp3", file_name_suffix=".md"
            ) as file_name_3:
                # Arrange
                scanner = MarkdownScanner()
                supplied_arguments = [
                    "--continue-on-error",
                    "--set",
                    extension_enable_front_matter,
                    "fix",
                    file_name_1,
                    file_name_2,
                    file_name_3,
                ]

                expected_return_code = 1
                expected_output = """Fixed: {file_name_1}
Fixed: {file_name_3}
""".replace(
                    "{file_name_1}", file_name_1
                ).replace(
                    "{file_name_3}", file_name_3
                )
                expected_error = """{file_name_2}:0:0: An unhandled error occurred processing the document.
""".replace(
                    "{file_name_2}", file_name_2
                )
                # Act
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

                # Assert
                execute_results.assert_results(
                    expected_output, expected_error, expected_return_code
                )



"""
Module to provide tests related to the basic parts of the scanner.
"""

import os
import tempfile
from test.markdown_scanner import MarkdownScanner

from .utils import create_temporary_configuration_file, temporary_change_to_directory


def test_markdown_with_dash_e_single_by_id_and_good_config():
    """
    Test to make sure we get enable a rule if '-e' is supplied and the id of the
    rule is provided. The test data for MD047 is used as it is a simple file that
    passes normally, it is used as a comparison.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"plugins": {"md999": {"test_value": 2}}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-e",
            "MD999",
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 0
        expected_output = """MD999>>init_from_config
MD999>>test_value>>2
MD999>>other_test_value>>1
MD999>>starting_new_file>>
MD999>>token:[atx(1,1):1:0:]
MD999>>token:[text(1,3):This is a test: ]
MD999>>token:[end-atx::]
MD999>>token:[BLANK(2,1):]
MD999>>token:[para(3,1):]
MD999>>token:[text(3,1):The line after this line should be blank.:]
MD999>>token:[end-para:::True]
MD999>>token:[BLANK(4,1):]
MD999>>token:[end-of-stream(5,0)]
MD999>>next_line:# This is a test
MD999>>next_line:
MD999>>next_line:The line after this line should be blank.
MD999>>next_line:
MD999>>completed_file
"""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_dash_e_single_by_id_and_bad_config():
    """
    Test to make sure we get enable a rule if '-e' is supplied and the id of the
    rule is provided. The test data for MD047 is used as it is a simple file that
    passes normally, it is used as a comparison.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"plugins": {"md999": {"test_value": "fred"}}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-e",
            "MD999",
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 0
        expected_output = """MD999>>init_from_config
MD999>>test_value>>1
MD999>>other_test_value>>1
MD999>>starting_new_file>>
MD999>>token:[atx(1,1):1:0:]
MD999>>token:[text(1,3):This is a test: ]
MD999>>token:[end-atx::]
MD999>>token:[BLANK(2,1):]
MD999>>token:[para(3,1):]
MD999>>token:[text(3,1):The line after this line should be blank.:]
MD999>>token:[end-para:::True]
MD999>>token:[BLANK(4,1):]
MD999>>token:[end-of-stream(5,0)]
MD999>>next_line:# This is a test
MD999>>next_line:
MD999>>next_line:The line after this line should be blank.
MD999>>next_line:
MD999>>completed_file
"""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_dash_e_single_by_id_and_bad_config_file():
    """
    Test to make sure we get an error if we provide a configuration file that is
    in a json format, but not valid.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"plugins": {"myrule.md999": {"test_value": "fred"}}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-e",
            "MD999",
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 1
        expected_output = ""
        expected_error = (
            "Specified configuration file '"
            + configuration_file
            + "' is not valid: Key strings cannot contain a whitespace character, a '=' character, or a '.' character.\n"
        )

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_dash_e_single_by_id_and_non_json_config_file():
    """
    Test to make sure we get an error if we provide a configuration file that is
    not in a json format.

    This function shadows
    test_api_config_with_bad_contents
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = "not a json file"
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-e",
            "MD999",
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 1
        expected_output = ""
        expected_error = f"Specified configuration file '{configuration_file}' was not parseable as a JSON, YAML, or TOML file."

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_dash_e_single_by_id_and_non_present_config_file():
    """
    Test to make sure we get an error if we provide a configuration file that is
    not in a json format.

    This function shadows
    test_api_config_with_bad_path
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    configuration_file = "not-exists"
    assert not os.path.exists(configuration_file)
    supplied_arguments = [
        "-e",
        "MD999",
        "-c",
        configuration_file,
        "scan",
        source_path,
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """

Specified configuration file `not-exists` does not exist."""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_e_single_by_id_and_good_select_config():
    """
    Test to make sure we get enable a rule if '-e' is supplied and the id of the
    rule is provided. The test data for MD047 is used as it is a simple file that
    passes normally, it is used as a comparison.

    This function shadows
    test_api_config_with_config_file_with_good_value
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"plugins": {"md999": {"other_test_value": 2}}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-e",
            "MD999",
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 0
        expected_output = """MD999>>init_from_config
MD999>>test_value>>1
MD999>>other_test_value>>2
MD999>>starting_new_file>>
MD999>>token:[atx(1,1):1:0:]
MD999>>token:[text(1,3):This is a test: ]
MD999>>token:[end-atx::]
MD999>>token:[BLANK(2,1):]
MD999>>token:[para(3,1):]
MD999>>token:[text(3,1):The line after this line should be blank.:]
MD999>>token:[end-para:::True]
MD999>>token:[BLANK(4,1):]
MD999>>token:[end-of-stream(5,0)]
MD999>>next_line:# This is a test
MD999>>next_line:
MD999>>next_line:The line after this line should be blank.
MD999>>next_line:
MD999>>completed_file
"""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_dash_e_single_by_id_and_bad_select_config():
    """
    Test to make sure we get enable a rule if '-e' is supplied and the id of the
    rule is provided. The test data for MD047 is used as it is a simple file that
    passes normally, it is used as a comparison.

    This function is shadowed by
    test_api_config_with_config_file_with_bad_value
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"plugins": {"MD999": {"other_test_value": 9}}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-e",
            "MD999",
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 0
        expected_output = """MD999>>init_from_config
MD999>>test_value>>1
MD999>>other_test_value>>1
MD999>>starting_new_file>>
MD999>>token:[atx(1,1):1:0:]
MD999>>token:[text(1,3):This is a test: ]
MD999>>token:[end-atx::]
MD999>>token:[BLANK(2,1):]
MD999>>token:[para(3,1):]
MD999>>token:[text(3,1):The line after this line should be blank.:]
MD999>>token:[end-para:::True]
MD999>>token:[BLANK(4,1):]
MD999>>token:[end-of-stream(5,0)]
MD999>>next_line:# This is a test
MD999>>next_line:
MD999>>next_line:The line after this line should be blank.
MD999>>next_line:
MD999>>completed_file
"""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_dash_e_single_by_id_and_config_causing_config_exception():
    """
    Test to make sure if we tell the test plugin to throw an exception during the
    call to `initialize_from_config`, that it is handled properly.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"plugins": {"md999": {"test_value": 10}}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-e",
            "MD999",
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 1
        expected_output = """MD999>>init_from_config
MD999>>test_value>>10
MD999>>other_test_value>>1
"""
        expected_error = """BadPluginError encountered while configuring plugins:
Plugin id 'MD999' had a critical failure during the '__apply_configuration' action.
"""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_dash_e_single_by_id_and_config_causing_next_token_exception():
    """
    Test to make sure if we tell the test plugin to throw an exception during the
    call to `next_token`, that it is handled properly.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"plugins": {"md999": {"test_value": 20}}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-e",
            "MD999",
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 1
        expected_output = """MD999>>init_from_config
MD999>>test_value>>20
MD999>>other_test_value>>1
MD999>>starting_new_file>>
MD999>>token:[atx(1,1):1:0:]
"""
        expected_error = """BadPluginError encountered while scanning '{source_path}':
(1,1): Plugin id 'MD999' had a critical failure during the 'next_token' action.
""".replace(
            "{source_path}", source_path
        )

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_bad_strict_config_type():
    """
    Test to make sure that we can set the strict configuration mode from
    the configuration file, capturing any bad errors.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"mode": {"strict-config": 2}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 1
        expected_output = ""
        expected_error = "Configuration Error: The value for property 'mode.strict-config' must be of type 'bool'."

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_good_strict_config_type():
    """
    Test to make sure that we can set the strict configuration mode from
    the configuration file, capturing any bad errors.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_configuration = {"mode": {"strict-config": True}, "log": {"file": 0}}
    with create_temporary_configuration_file(
        supplied_configuration
    ) as configuration_file:
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            source_path,
        ]

        expected_return_code = 1
        expected_output = ""
        expected_error = "Configuration Error: The value for property 'log.file' must be of type 'str'.\n"

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_with_default_configuration_file_with_error():
    """
    Test to make sure that a default configuration will be read and have the
    same errors as if it was specified on the command line.

    This function shadows
    test_api_config_with_bad_contents_for_default_config
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    default_configuration = {"mode": {"strict-config": True}, "log": {"file": 0}}

    # Note that the default configuration file is determined by the current working
    # directory, so this test creates a new directory and executes the parser from
    # within that directory.
    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            default_configuration, file_name=".pymarkdown", directory=tmp_dir_path
        ):
            supplied_arguments = [
                "scan",
                source_path,
            ]

            expected_return_code = 1
            expected_output = ""
            expected_error = "Configuration Error: The value for property 'log.file' must be of type 'str'.\n"

            # Act
            with temporary_change_to_directory(tmp_dir_path):
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

            # Assert
            execute_results.assert_results(
                expected_output, expected_error, expected_return_code
            )


def test_markdown_with_overlapping_configuration_files():
    """
    Test to make sure that information from a default configuration file and
    a specified configuration file give each other the right layering.
    i.e. specific config file should override default config file
    """

    # Arrange
    scanner = MarkdownScanner()
    xxx = """# This is a document

* a list
  - a sublist
  - a very long sublist item

this is a very long line
"""

    # The `ul-style` item is set by the default configuration and overridden as "asterisk" by the supplied configuration.
    # The `line-length` item is set by the default configuration as `10` and is not overridden.
    # This results in a `ul-style` of `asterisk` and a `line-length` of `10`.
    default_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"ul-style": {"style": "dash"}, "line-length": {"line_length": 10}},
    }
    supplied_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"ul-style": {"style": "asterisk"}},
    }

    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            supplied_configuration, directory=tmp_dir_path
        ) as configuration_file:
            with create_temporary_configuration_file(
                default_configuration, file_name=".pymarkdown", directory=tmp_dir_path
            ):
                supplied_arguments = ["-c", configuration_file, "scan-stdin"]

                expected_return_code = 1
                expected_output = (
                    "stdin:4:3: MD004: Inconsistent Unordered List Start style "
                    + "[Expected: asterisk; Actual: dash] (ul-style)\n"
                    + "stdin:5:1: MD013: Line length [Expected: 10, Actual: 28] (line-length)\n"
                    + "stdin:7:1: MD013: Line length [Expected: 10, Actual: 24] (line-length)"
                )
                expected_error = ""

                # Act
                with temporary_change_to_directory(tmp_dir_path):
                    execute_results = scanner.invoke_main(
                        arguments=supplied_arguments, standard_input_to_use=xxx
                    )

                # Assert
                execute_results.assert_results(
                    expected_output, expected_error, expected_return_code
                )


def test_markdown_with_pyproject_configuration_file_with_error():
    """
    Test to make sure that a pyproject configuration will be read and have the
    same errors as if it was specified on the command line.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    default_configuration = """
[tool.pymarkdown]
log.file = 2
a.c = "3"
"""

    # Note that the default configuration file is determined by the current working
    # directory, so this test creates a new directory and executes the parser from
    # within that directory.
    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            default_configuration, file_name="pyproject.toml", directory=tmp_dir_path
        ):
            supplied_arguments = [
                "--strict-config",
                "scan",
                source_path,
            ]

            expected_return_code = 1
            expected_output = ""
            expected_error = "Configuration Error: The value for property 'log.file' must be of type 'str'.\n"

            # Act
            with temporary_change_to_directory(tmp_dir_path):
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

            # Assert
            execute_results.assert_results(
                expected_output, expected_error, expected_return_code
            )


def test_markdown_with_pyproject_direct_configuration_file_with_error():
    """
    Test to make sure that a pyproject configuration will be read and have the
    same errors as if it was specified on the command line.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    default_configuration = """
[tool.pymarkdown]
plugins.MD013.enabled = false
plugins.MD003.style = "atx"
plugins.MD004.style = "asterisk"
plugins.MD007.indent = 4
plugins.MD033.enabled = false
plugins.MD041.enabled = false
plugins.MD022.enabled = false
log.file = 2
"""

    # Note that the default configuration file is determined by the current working
    # directory, so this test creates a new directory and executes the parser from
    # within that directory.
    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            default_configuration,
            file_name="pyproject.toml",
            directory=tmp_dir_path,
            file_name_suffix=".toml",
        ) as config_path:
            supplied_arguments = [
                "--strict-config",
                "--config",
                config_path,
                "scan",
                source_path,
            ]

            expected_return_code = 1
            expected_output = ""
            expected_error = "Configuration Error: The value for property 'log.file' must be of type 'str'.\n"

            # Act
            with temporary_change_to_directory(tmp_dir_path):
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

            # Assert
            execute_results.assert_results(
                expected_output, expected_error, expected_return_code
            )



"""
Module to provide tests for the application file scanner module.
"""

import argparse
import sys
from test.utils import assert_that_exception_is_raised, compare_expected_to_actual

from pymarkdown.application_file_scanner import ApplicationFileScanner

if sys.version_info < (3, 10):
    ARGPARSE_X = "optional arguments:"
else:
    ARGPARSE_X = "options:"
if sys.version_info < (3, 13):
    ALT_EXTENSIONS_X = (
        "-ae ALTERNATE_EXTENSIONS, --alternate-extensions ALTERNATE_EXTENSIONS"
    )
    EXCLUSIONS_X = "-e PATH_EXCLUSIONS, --exclude PATH_EXCLUSIONS"
else:
    ALT_EXTENSIONS_X = "-ae, --alternate-extensions ALTERNATE_EXTENSIONS"
    EXCLUSIONS_X = "-e, --exclude PATH_EXCLUSIONS"


def test_application_file_scanner_args_no_changes():
    """
    Test to make sure we get all scanner args without any flags changed.
    """

    # Arrange
    expected_output = f"""usage: pytest [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS] [-e PATH_EXCLUSIONS]
              path [path ...]

Lint any found files.

positional arguments:
  path                  one or more paths to examine for eligible files

{ARGPARSE_X}
  -h, --help            show this help message and exit
  -l, --list-files      list any eligible files found on the specified paths
                        and exit
  -r, --recurse         recursively traverse any found directories for
                        matching files
  {ALT_EXTENSIONS_X}
                        provide an alternate set of file extensions to match
                        against
  {EXCLUSIONS_X}
                        one or more paths to exclude from the search. Can be a
                        glob pattern."""
    parser = argparse.ArgumentParser(description="Lint any found files.", prog="pytest")

    # Act
    ApplicationFileScanner.add_default_command_line_arguments(parser, ".md")
    args = parser.format_help()

    # Assert
    compare_expected_to_actual(expected_output, args)


def test_application_file_scanner_args_bad_extension():
    """
    Test to make sure we get all scanner args with a bad default extension.
    """

    # Arrange
    expected_output = "Extension '*.md' is not a valid extension: Extension '*.md' must start with a period."
    parser = argparse.ArgumentParser(description="Lint any found files.", prog="pytest")

    # Act & Assert
    assert_that_exception_is_raised(
        argparse.ArgumentTypeError,
        expected_output,
        ApplicationFileScanner.add_default_command_line_arguments,
        parser,
        "*.md",
    )


def test_application_file_scanner_args_with_file_type_name():
    """
    Test to make sure we get all scanner args with a file type name specified.
    """

    # Arrange
    expected_output = f"""usage: pytest [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS] [-e PATH_EXCLUSIONS]
              path [path ...]

Lint any found files.

positional arguments:
  path                  one or more paths to examine for eligible MINE files

{ARGPARSE_X}
  -h, --help            show this help message and exit
  -l, --list-files      list any eligible MINE files found on the specified
                        paths and exit
  -r, --recurse         recursively traverse any found directories for
                        matching files
  {ALT_EXTENSIONS_X}
                        provide an alternate set of file extensions to match
                        against
  {EXCLUSIONS_X}
                        one or more paths to exclude from the search. Can be a
                        glob pattern."""
    parser = argparse.ArgumentParser(description="Lint any found files.", prog="pytest")

    # Act
    ApplicationFileScanner.add_default_command_line_arguments(
        parser, ".md", file_type_name="MINE"
    )
    args = parser.format_help()

    # Assert
    compare_expected_to_actual(expected_output, args)


def test_application_file_scanner_args_with_empty_file_type_name():
    """
    Test to make sure we get all scanner args with an empty file type name specified.
    """

    # Arrange
    expected_output = f"""usage: pytest [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS] [-e PATH_EXCLUSIONS]
              path [path ...]

Lint any found files.

positional arguments:
  path                  one or more paths to examine for eligible files

{ARGPARSE_X}
  -h, --help            show this help message and exit
  -l, --list-files      list any eligible files found on the specified paths
                        and exit
  -r, --recurse         recursively traverse any found directories for
                        matching files
  {ALT_EXTENSIONS_X}
                        provide an alternate set of file extensions to match
                        against
  {EXCLUSIONS_X}
                        one or more paths to exclude from the search. Can be a
                        glob pattern."""
    parser = argparse.ArgumentParser(description="Lint any found files.", prog="pytest")

    # Act
    ApplicationFileScanner.add_default_command_line_arguments(
        parser, ".md", file_type_name=""
    )
    args = parser.format_help()

    # Assert
    compare_expected_to_actual(expected_output, args)


def test_application_file_scanner_args_without_list_files():
    """
    Test to make sure we get all scanner args with list files disabled
    """

    # Arrange
    expected_output = f"""usage: pytest [-h] [-r] [-ae ALTERNATE_EXTENSIONS] [-e PATH_EXCLUSIONS]
              path [path ...]

Lint any found files.

positional arguments:
  path                  one or more paths to examine for eligible files

{ARGPARSE_X}
  -h, --help            show this help message and exit
  -r, --recurse         recursively traverse any found directories for
                        matching files
  {ALT_EXTENSIONS_X}
                        provide an alternate set of file extensions to match
                        against
  {EXCLUSIONS_X}
                        one or more paths to exclude from the search. Can be a
                        glob pattern."""
    parser = argparse.ArgumentParser(description="Lint any found files.", prog="pytest")

    # Act
    ApplicationFileScanner.add_default_command_line_arguments(
        parser, ".md", show_list_files=False
    )
    args = parser.format_help()

    # Assert
    compare_expected_to_actual(expected_output, args)


def test_application_file_scanner_args_without_recurse_directories():
    """
    Test to make sure we get all scanner args with recurse directories disabled
    """

    # Arrange
    expected_output = f"""usage: pytest [-h] [-l] [-ae ALTERNATE_EXTENSIONS] [-e PATH_EXCLUSIONS]
              path [path ...]

Lint any found files.

positional arguments:
  path                  one or more paths to examine for eligible files

{ARGPARSE_X}
  -h, --help            show this help message and exit
  -l, --list-files      list any eligible files found on the specified paths
                        and exit
  {ALT_EXTENSIONS_X}
                        provide an alternate set of file extensions to match
                        against
  {EXCLUSIONS_X}
                        one or more paths to exclude from the search. Can be a
                        glob pattern."""
    parser = argparse.ArgumentParser(description="Lint any found files.", prog="pytest")

    # Act
    ApplicationFileScanner.add_default_command_line_arguments(
        parser, ".md", show_recurse_directories=False
    )
    args = parser.format_help()

    # Assert
    compare_expected_to_actual(expected_output, args)


def test_application_file_scanner_args_without_alternate_extensions():
    """
    Test to make sure we get all scanner args with alternate extensions disabled
    """

    # Arrange
    expected_output = f"""usage: pytest [-h] [-l] [-r] [-e PATH_EXCLUSIONS] path [path ...]

Lint any found files.

positional arguments:
  path                  one or more paths to examine for eligible files

{ARGPARSE_X}
  -h, --help            show this help message and exit
  -l, --list-files      list any eligible files found on the specified paths
                        and exit
  -r, --recurse         recursively traverse any found directories for
                        matching files
  {EXCLUSIONS_X}
                        one or more paths to exclude from the search. Can be a
                        glob pattern."""
    parser = argparse.ArgumentParser(description="Lint any found files.", prog="pytest")

    # Act
    ApplicationFileScanner.add_default_command_line_arguments(
        parser, ".md", show_alternate_extensions=False
    )
    args = parser.format_help()

    # Assert
    compare_expected_to_actual(expected_output, args)


def test_application_file_scanner_args_without_exclusions():
    """
    Test to make sure we get all scanner args with alternate extensions disabled
    """

    # Arrange
    expected_output = f"""usage: pytest [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS] path [path ...]

Lint any found files.

positional arguments:
  path                  one or more paths to examine for eligible files

{ARGPARSE_X}
  -h, --help            show this help message and exit
  -l, --list-files      list any eligible files found on the specified paths
                        and exit
  -r, --recurse         recursively traverse any found directories for
                        matching files
  {ALT_EXTENSIONS_X}
                        provide an alternate set of file extensions to match
                        against"""
    parser = argparse.ArgumentParser(description="Lint any found files.", prog="pytest")

    # Act
    ApplicationFileScanner.add_default_command_line_arguments(
        parser, ".md", show_exclusions=False
    )
    args = parser.format_help()

    # Assert
    compare_expected_to_actual(expected_output, args)



"""
Module to provide tests related to the basic parts of the scanner.
"""

import os
import tempfile
from test.markdown_scanner import MarkdownScanner

import yaml

from .utils import create_temporary_configuration_file, temporary_change_to_directory

__TEST_DOCUMENT = """# This is a document

* a list
  - a sublist
  - a very long sublist item

this is a very long line
"""


def test_markdown_with_config_no_config():
    """
    Test to make sure that we have a baseline for the other configuration tests.

    This function shadows
    test_api_scan_string_test_bad_file_due_to_no_disables
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    supplied_arguments = [
        "--strict-config",
        "scan-stdin",
    ]

    expected_return_code = 1
    expected_output = "stdin:4:3: MD004: Inconsistent Unordered List Start style [Expected: asterisk; Actual: dash] (ul-style)"
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, standard_input_to_use=stdin_to_use
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_config_general_command_line():
    """
    Test to make sure that we can disable a rule from the command line without
    any configuration file.

    This test shadows
    test_api_scan_string_test_good_file_after_disables
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    supplied_arguments = [
        "--strict-config",
        "-d",
        "Md004",
        "scan-stdin",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, standard_input_to_use=stdin_to_use
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_config_general_command_line_and_specific_command_line():
    """
    Test to make sure that we can disable a rule specifically and enable a
    rule through a configuration setting, with the disable winning as it is
    the more specific configuration.
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    supplied_arguments = [
        "--strict-config",
        "-d",
        "Md004",
        "-s",
        "plugins.md004.enabled=$!True",
        "scan-stdin",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, standard_input_to_use=stdin_to_use
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_config_json_configuration_file():
    """
    Test to make sure that we can supply configuration via a JSON configuration file.
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    specified_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"md004": {"enabled": False}},
    }

    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            specified_configuration, file_name="myconfig", directory=tmp_dir_path
        ) as configuration_file:
            supplied_arguments = [
                "-c",
                configuration_file,
                "scan-stdin",
            ]

            expected_return_code = 0
            expected_output = ""
            expected_error = ""

            # Act
            execute_results = scanner.invoke_main(
                arguments=supplied_arguments, standard_input_to_use=stdin_to_use
            )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_config_yaml_configuration_file():
    """
    Test to make sure that we can supply configuration via a YAML configuration file.
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    specified_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"md004": {"enabled": False}},
    }

    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            yaml.dump(specified_configuration),
            file_name="myconfig",
            directory=tmp_dir_path,
        ) as configuration_file:
            supplied_arguments = [
                "--stack-trace",
                "-c",
                configuration_file,
                "scan-stdin",
            ]

            expected_return_code = 0
            expected_output = ""
            expected_error = ""

            # Act
            execute_results = scanner.invoke_main(
                arguments=supplied_arguments, standard_input_to_use=stdin_to_use
            )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_config_default_yaml():
    """
    Test to make sure that we can supply configuration via a YAML configuration file.
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    specified_yaml_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"md004": {"enabled": False}},
    }

    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            yaml.dump(specified_yaml_configuration),
            file_name=".pymarkdown.yaml",
            directory=tmp_dir_path,
        ):
            supplied_arguments = [
                "scan-stdin",
            ]

            expected_return_code = 0
            expected_output = ""
            expected_error = ""

            # Act
            with temporary_change_to_directory(tmp_dir_path):
                execute_results = scanner.invoke_main(
                    arguments=supplied_arguments, standard_input_to_use=stdin_to_use
                )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_config_default_yml():
    """
    Test to make sure that we can supply configuration via a YML configuration file.
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    specified_yaml_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"md004": {"enabled": False}},
    }

    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            yaml.dump(specified_yaml_configuration),
            file_name=".pymarkdown.yaml",
            directory=tmp_dir_path,
        ):
            supplied_arguments = [
                "scan-stdin",
            ]

            expected_return_code = 0
            expected_output = ""
            expected_error = ""

            # Act
            with temporary_change_to_directory(tmp_dir_path):
                execute_results = scanner.invoke_main(
                    arguments=supplied_arguments, standard_input_to_use=stdin_to_use
                )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_config_default_json_beats_default_yaml():
    """
    Test to make sure that we can supply configuration via a YAML configuration file.
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    specified_json_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"md004": {"enabled": False}},
    }
    specified_yaml_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"md004": {"enabled": True}},
    }

    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            specified_json_configuration,
            file_name=".pymarkdown",
            directory=tmp_dir_path,
        ):
            with create_temporary_configuration_file(
                yaml.dump(specified_yaml_configuration),
                file_name=".pymarkdown.yaml",
                directory=tmp_dir_path,
            ):
                supplied_arguments = [
                    "scan-stdin",
                ]

                expected_return_code = 0
                expected_output = ""
                expected_error = ""

                # Act
                with temporary_change_to_directory(tmp_dir_path):
                    execute_results = scanner.invoke_main(
                        arguments=supplied_arguments, standard_input_to_use=stdin_to_use
                    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_config_bad_json_and_yaml_configuration_file():
    """
    Test to make sure that we error out properly with a file that is
    not JSON and is not YAML.xxx
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    specified_configuration = """hallo: 1
bye
"""
    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            specified_configuration, file_name="myconfig", directory=tmp_dir_path
        ) as configuration_file:
            supplied_arguments = [
                "--stack-trace",
                "-c",
                configuration_file,
                "scan-stdin",
            ]

            expected_return_code = 1
            expected_output = ""
            expected_error = f"Specified configuration file '{configuration_file}' was not parseable as a JSON, YAML, or TOML file."

            # Act
            execute_results = scanner.invoke_main(
                arguments=supplied_arguments, standard_input_to_use=stdin_to_use
            )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_config_specific_command_line_and_configuration_file():
    """
    Test to make sure that if a command line setting and a configuration
    file setting are applied, that the command line setting wins as it is
    more specific.
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    specified_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"md004": {"enabled": True}},
    }

    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            specified_configuration, file_name="myconfig", directory=tmp_dir_path
        ) as configuration_file:
            supplied_arguments = [
                "--strict-config",
                "-s",
                "plugins.md004.enabled=$!False",
                "-c",
                configuration_file,
                "scan-stdin",
            ]

            expected_return_code = 0
            expected_output = ""
            expected_error = ""

            # Act
            execute_results = scanner.invoke_main(
                arguments=supplied_arguments, standard_input_to_use=stdin_to_use
            )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_config_configuration_file_and_default_configuration_file():
    """
    Test to make sure that any default configuration is overridden by an explicitly
    named configuration file.
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    specified_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"md004": {"enabled": False}},
    }
    default_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"md004": {"enabled": True}},
    }

    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            specified_configuration, file_name="myconfig", directory=tmp_dir_path
        ) as configuration_file:
            with create_temporary_configuration_file(
                default_configuration, file_name=".pymarkdown", directory=tmp_dir_path
            ):
                supplied_arguments = [
                    "--strict-config",
                    "-c",
                    configuration_file,
                    "scan-stdin",
                ]

                expected_return_code = 0
                expected_output = ""
                expected_error = ""

                # Act
                with temporary_change_to_directory(tmp_dir_path):
                    execute_results = scanner.invoke_main(
                        arguments=supplied_arguments, standard_input_to_use=stdin_to_use
                    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_default_configuration_file_and_alternate_configuration_file():
    """
    Test to make sure that a default configuration file takes precedence over a
    Python standard configuration file.
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    default_configuration = {
        "mode": {"strict-config": True},
        "plugins": {"md004": {"enabled": False}},
    }
    alternate_configuration = """
[tool.pymarkdown]
plugins.md004.enabled = true
"""

    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            default_configuration, file_name=".pymarkdown", directory=tmp_dir_path
        ):
            with create_temporary_configuration_file(
                alternate_configuration,
                file_name="pyproject.toml",
                directory=tmp_dir_path,
            ):
                supplied_arguments = [
                    "--strict-config",
                    "scan-stdin",
                ]

                expected_return_code = 0
                expected_output = ""
                expected_error = ""

                # Act
                with temporary_change_to_directory(tmp_dir_path):
                    execute_results = scanner.invoke_main(
                        arguments=supplied_arguments, standard_input_to_use=stdin_to_use
                    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_alternate_configuration_file():
    """
    Test to make sure that we can load configuration from a Python standard
    configuration file.
    """

    # Arrange
    scanner = MarkdownScanner()
    stdin_to_use = __TEST_DOCUMENT
    alternate_configuration = """
[tool.pymarkdown]
plugins.md004.enabled = false
"""

    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            alternate_configuration, file_name="pyproject.toml", directory=tmp_dir_path
        ):
            supplied_arguments = [
                "--strict-config",
                "scan-stdin",
            ]

            expected_return_code = 0
            expected_output = ""
            expected_error = ""

            # Act
            with temporary_change_to_directory(tmp_dir_path):
                execute_results = scanner.invoke_main(
                    arguments=supplied_arguments, standard_input_to_use=stdin_to_use
                )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_pyproject_configuration_file_with_bad_log_file_value_type():
    """
    Test to make sure that a standard python configuration file can be used and
    that an error with value types is reported.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    default_configuration = """
[tool.pymarkdown]
log.file = 2
a.c = "3"
"""

    # Note that the default configuration file is determined by the current working
    # directory, so this test creates a new directory and executes the parser from
    # within that directory.
    with tempfile.TemporaryDirectory() as tmp_dir_path:
        with create_temporary_configuration_file(
            default_configuration, file_name="pyproject.toml", directory=tmp_dir_path
        ):
            supplied_arguments = [
                "--strict-config",
                "scan",
                source_path,
            ]

            expected_return_code = 1
            expected_output = ""
            expected_error = "Configuration Error: The value for property 'log.file' must be of type 'str'.\n"

            # Act
            with temporary_change_to_directory(tmp_dir_path):
                execute_results = scanner.invoke_main(arguments=supplied_arguments)

            # Assert
            execute_results.assert_results(
                expected_output, expected_error, expected_return_code
            )



"""
Module to provide tests related to logging.
"""

import logging
import os
import tempfile
from test.markdown_scanner import MarkdownScanner

from pymarkdown.application_logging import ApplicationLogging
from pymarkdown.container_blocks.container_indices import ContainerIndices
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.tokens.stack_token import StackToken

from .utils import (
    assert_file_is_as_expected,
    create_temporary_configuration_file,
    create_temporary_file_for_reuse,
    read_contents_of_text_file,
)

POGGER = ParserLogger(logging.getLogger(__name__))


def test_markdown_with_dash_dash_log_level_debug(caplog):
    """
    Test to make sure we get the right effect if the `--log-level` flag
    is set for debug.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "--log-level",
        "DEBUG",
        "scan",
        source_path,
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = """"""

    # Act
    ParserLogger.sync_on_next_call()
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )

    # Info messages
    assert "Number of files found: " in caplog.text
    assert f"Determining files to scan for path '{source_path}'." in caplog.text

    # Debug messages
    assert f"Provided path '{source_path}' is a valid file. Adding." in caplog.text


def test_markdown_with_dash_dash_log_level_info(caplog):
    """
    Test to make sure we get the right effect if the `--log-level` flag
    is set for info.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "--log-level",
        "INFO",
        "scan",
        source_path,
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    ParserLogger.sync_on_next_call()
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )

    # Info messages
    assert "Number of files found: " in caplog.text
    assert f"Determining files to scan for path '{source_path}'." in caplog.text

    # Debug messages
    assert f"Provided path '{source_path}' is a valid file. Adding." not in caplog.text


def test_markdown_with_dash_dash_log_level_invalid(caplog):
    """
    Test to make sure we get the right effect if the `--log-level` flag
    is set for an invalid log level.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "--log-level",
        "invalid",
        "scan",
        source_path,
    ]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py [-h] [-e ENABLE_RULES] [-d DISABLE_RULES]
               [--add-plugin ADD_PLUGIN] [--config CONFIGURATION_FILE]
               [--set SET_CONFIGURATION] [--strict-config] [--stack-trace]
               [--continue-on-error]
               [--log-level {CRITICAL,ERROR,WARNING,INFO,DEBUG}]
               [--log-file LOG_FILE] [--return-code-scheme {default,minimal}]
               {extensions,fix,plugins,scan,scan-stdin,version} ...
main.py: error: argument --log-level: invalid validate_log_level_type value: 'invalid'
"""

    # Act
    ParserLogger.sync_on_next_call()
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )

    # Info messages
    assert "Number of scanned files found: " not in caplog.text
    assert (
        "Determining files to scan for path "
        + "'test/resources/rules/md047/end_with_blank_line.md'."
        not in caplog.text
    )

    # Debug messages
    assert (
        "Provided path 'test/resources/rules/md047/end_with_blank_line.md' "
        + "is a valid Markdown file. Adding."
        not in caplog.text
    )


def test_markdown_with_dash_dash_log_level_info_with_file():
    # sourcery skip: extract-method
    """
    Test to make sure we get the right effect if the `--log-level` flag
    is set for info with the results going to a file.

    This function is shadowed in test_api_simple_clean_scan.
    """

    # Arrange
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    with create_temporary_file_for_reuse() as log_file_name:
        scanner = MarkdownScanner()
        supplied_arguments = [
            "--log-level",
            "INFO",
            "--log-file",
            log_file_name,
            "scan",
            source_path,
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        ParserLogger.sync_on_next_call()
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )

        file_data = read_contents_of_text_file(log_file_name).replace("\n", "")

        # Info messages
        assert "Number of files found: " in file_data, f">{file_data}<"
        assert f"Determining files to scan for path '{source_path}'." in file_data

        # Debug messages
        assert (
            f"Provided path '{source_path}' is a valid file. Adding." not in file_data
        )


def test_markdown_with_dash_dash_log_level_info_with_file_as_directory():
    """
    Test to make sure we get the right effect if a directory is specified
    as the file to log to.

    This is shadowed by
    test_api_logging_debug_to_file_as_directory
    """

    # Arrange
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    with tempfile.TemporaryDirectory() as temp_directory:
        log_file_name = temp_directory

        scanner = MarkdownScanner()
        supplied_arguments = [
            "--log-file",
            log_file_name,
            "scan",
            source_path,
        ]

        expected_return_code = 1
        expected_output = ""
        expected_error = "Unexpected Error(ApplicationLoggingException): Failure initializing logging subsystem."

        # Act
        ParserLogger.sync_on_next_call()
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )


def test_markdown_logger_stack_token_normal_output():
    """
    Test for normal output of the stack token for debug output.
    """

    # Arrange
    stack_token = StackToken("type", extra_data=None)

    # Act
    stack_token_string = str(stack_token)

    # Assert
    assert stack_token_string == "StackToken(type)"


def test_markdown_logger_stack_token_extra_output():
    """
    Test for extra output of the stack token for debug output.
    """

    # Arrange
    stack_token = StackToken("type", extra_data="abc")

    # Act
    stack_token_string = str(stack_token)

    # Assert
    assert stack_token_string == "StackToken(type:abc)"


def test_markdown_logger_container_indices():
    """
    Test for ContainerIndices class for output
    """

    # Arrange
    container_indices = ContainerIndices(1, 2, 3)

    # Act
    container_indices_string = str(container_indices)

    # Assert
    assert (
        container_indices_string
        == "{ContainerIndices:ulist_index:1;olist_index:2;block_index:3}"
    )


def test_markdown_logger_active():
    """
    Since calls to it are commented out most of the time,
    test the debug_with_visible_whitespace function manually.
    """
    configuration_file_name = None
    new_handler = None
    root_logger = None
    file_as_lines = None
    with create_temporary_configuration_file("") as configuration_file_name:
        try:
            new_handler = logging.FileHandler(configuration_file_name)

            root_logger = logging.getLogger()
            root_logger.setLevel(logging.DEBUG)
            root_logger.addHandler(new_handler)

            assert POGGER.is_enabled_for(logging.DEBUG)
            new_logger = ParserLogger(logging.getLogger(__name__))
            new_logger.debug("simple logging")
        finally:
            if root_logger and new_handler:
                root_logger.removeHandler(new_handler)
                new_handler.close()

            file_as_lines = read_contents_of_text_file(configuration_file_name).split(
                "\n"
            )

    assert len(file_as_lines) == 2
    assert file_as_lines[0] == "simple logging"
    assert file_as_lines[1] == ""


def test_markdown_logger_inactive():
    """
    Since calls to it are commented out most of the time,
    test the debug_with_visible_whitespace function manually.
    """
    configuration_file_name = None
    new_handler = None
    root_logger = None
    with create_temporary_configuration_file("") as configuration_file_name:
        try:
            new_handler = logging.FileHandler(configuration_file_name)

            root_logger = logging.getLogger()
            root_logger.setLevel(logging.WARNING)
            root_logger.addHandler(new_handler)

            assert POGGER.is_enabled_for(logging.WARNING)
            new_logger = ParserLogger(logging.getLogger(__name__))
            new_logger.debug("simple logging")
        finally:
            if root_logger and new_handler:
                root_logger.removeHandler(new_handler)
                new_handler.close()
            file_contents = read_contents_of_text_file(configuration_file_name)

    assert not file_contents


def test_markdown_logger_arg_list_not_out_of_sync():
    """
    Since calls to it are commented out most of the time,
    test the debug_with_visible_whitespace function manually.
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.WARNING)
    assert POGGER.is_enabled_for(logging.WARNING)
    new_logger = ParserLogger(logging.getLogger(__name__))
    new_logger.debug_with_visible_whitespace("one sub $ and one in list", " 1 ")


def test_markdown_logger_arg_list_not_out_of_sync_with_debug():
    """
    Since calls to it are commented out most of the time,
    test the debug_with_visible_whitespace function manually.
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    assert POGGER.is_enabled_for(logging.DEBUG)
    new_logger = ParserLogger(logging.getLogger(__name__))
    new_logger.debug_with_visible_whitespace("one sub $ and one in list", " 1 ")


# pylint: disable=broad-exception-caught
def test_markdown_logger_arg_list_out_of_sync():
    """
    Test to verify that if we don't have the same number of
    $ characters in the string as arguments in the list, an
    exception with be thrown.
    """

    # Arrange
    try:
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)
        assert POGGER.is_enabled_for(logging.DEBUG)
        new_logger = ParserLogger(logging.getLogger(__name__))
        new_logger.debug("one sub $ but two in list", 1, 2)
        raise AssertionError("An exception should have been thrown by now.")
    except Exception as this_exception:
        assert (
            str(this_exception)
            == "The number of $ substitution characters does not equal the number of arguments in the list."
        )


# pylint: enable=broad-exception-caught


def test_markdown_logger_translate_log_level_valid():
    """
    Test to make sure that we can translate a valid logging level to its numeric counterpart.
    """

    # Arrange
    log_level_as_string = "DEBUG"

    # Act
    log_level = ApplicationLogging.translate_log_level(log_level_as_string)

    # Assert
    assert log_level == logging.DEBUG


def test_markdown_logger_translate_log_level_invalid():
    """
    Test to make sure that we can translate an invalid logging level to NOTSET
    to indicate it was bad.
    """

    # Arrange
    log_level_as_string = "other"

    # Act
    log_level = ApplicationLogging.translate_log_level(log_level_as_string)

    # Assert
    assert log_level == logging.NOTSET


def test_markdown_fix_with_no_rescan_log_debug(caplog):
    """
    Test that is a mirror of test_md005_bad_unordered_list_single_level_fix with
    slight changes to test logging.
    """

    # Arrange
    scanner = MarkdownScanner()
    original_file_contents = """* Item 1
 * Item 2
"""
    with create_temporary_configuration_file(
        original_file_contents, file_name_suffix=".md"
    ) as temp_source_path:
        supplied_arguments = [
            "--disable-rules",
            "md007,md029",
            "--log-level",
            "DEBUG",
            "-x-fix-no-rescan-log",
            "fix",
            temp_source_path,
        ]

        expected_return_code = 3
        expected_output = f"Fixed: {temp_source_path}"
        expected_error = ""

        expected_file_contents = """* Item 1
* Item 2
"""

        # Act
        ParserLogger.sync_on_next_call()
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
        assert_file_is_as_expected(temp_source_path, expected_file_contents)

        # Info messages
        assert "Number of files found: " in caplog.text
        assert (
            f"Determining files to scan for path '{temp_source_path}'." in caplog.text
        )

        # Debug messages
        assert (
            f"Provided path '{temp_source_path}' is a valid file. Adding."
            in caplog.text
        )

        # Warning messages
        assert (
            "Setting logging level to WARN during rescan upon request." in caplog.text
        )
        assert "Restoring log level to DEBUG." in caplog.text


def test_markdown_fix_with_no_rescan_log_info(caplog):
    """
    Test that is a mirror of test_md005_bad_unordered_list_single_level_fix with
    slight changes to test logging.
    """

    # Arrange
    scanner = MarkdownScanner()
    original_file_contents = """* Item 1
 * Item 2
"""
    with create_temporary_configuration_file(
        original_file_contents, file_name_suffix=".md"
    ) as temp_source_path:
        supplied_arguments = [
            "--disable-rules",
            "md007,md029",
            "--log-level",
            "INFO",
            "-x-fix-no-rescan-log",
            "fix",
            temp_source_path,
        ]

        expected_return_code = 3
        expected_output = f"Fixed: {temp_source_path}"
        expected_error = ""

        expected_file_contents = """* Item 1
* Item 2
"""

        # Act
        ParserLogger.sync_on_next_call()
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
        assert_file_is_as_expected(temp_source_path, expected_file_contents)

        # Info messages
        assert "Number of files found: " in caplog.text
        assert (
            f"Determining files to scan for path '{temp_source_path}'." in caplog.text
        )

        # Debug messages
        # assert f"Provided path '{temp_source_path}' is a valid file. Adding." in caplog.text

        # Warning messages
        assert (
            "Setting logging level to WARN during rescan upon request." in caplog.text
        )
        assert "Restoring log level to INFO." in caplog.text


def test_markdown_fix_with_no_rescan_log_warn(caplog):
    """
    Test that is a mirror of test_md005_bad_unordered_list_single_level_fix with
    slight changes to test logging.
    """

    # Arrange
    scanner = MarkdownScanner()
    original_file_contents = """* Item 1
 * Item 2
"""
    with create_temporary_configuration_file(
        original_file_contents, file_name_suffix=".md"
    ) as temp_source_path:
        supplied_arguments = [
            "--disable-rules",
            "md007,md029",
            "--log-level",
            "WARNING",
            "-x-fix-no-rescan-log",
            "fix",
            temp_source_path,
        ]

        expected_return_code = 3
        expected_output = f"Fixed: {temp_source_path}"
        expected_error = ""

        expected_file_contents = """* Item 1
* Item 2
"""

        # Act
        ParserLogger.sync_on_next_call()
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
        assert_file_is_as_expected(temp_source_path, expected_file_contents)

        # Info messages
        # assert "Number of files found: " in caplog.text
        # assert f"Determining files to scan for path '{temp_source_path}'." in caplog.text

        # Debug messages
        # assert f"Provided path '{temp_source_path}' is a valid file. Adding." in caplog.text

        # Warning messages
        assert (
            "Setting logging level to WARN during rescan upon request." in caplog.text
        )
        assert "Restoring log level to WARNING." in caplog.text



from typing import Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.leaf_blocks.table_block_tuple import TableRow
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class TableMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the main table markdown token.
    """

    def __init__(
        self,
        position_marker: PositionMarker,
    ) -> None:
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_main,
            extra_data=None,
            position_marker=position_marker,
            extracted_whitespace="",
            requires_end_token=True,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_main

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownToken,
            TableMarkdownToken.__rehydrate_table_start,
            TableMarkdownToken.__rehydrate_table_end,
        )

    @staticmethod
    def __rehydrate_table_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return ""

    @staticmethod
    def __rehydrate_table_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return ""

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownToken,
            TableMarkdownToken.__handle_start_table_token,
            TableMarkdownToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        token_parts = []
        if (
            not output_html
            and transform_state.transform_stack
            and transform_state.transform_stack[-1].endswith("<li>")
        ):
            token_parts.append(ParserHelper.newline_character)
        token_parts.extend([output_html, "<table>", ParserHelper.newline_character])
        return "".join(token_parts)

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        return "".join([output_html, "</table>", ParserHelper.newline_character])


class TableMarkdownHeaderToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the table header markdown token.
    """

    def __init__(
        self,
        header_table_row: TableRow,
        separator_table_row: TableRow,
        position_marker: PositionMarker,
    ) -> None:
        if header_table_row:
            self.header_row_leading_whitespace = header_table_row.extracted_whitespace
            self.header_row_trailing_whitespace = header_table_row.trailing_whitespace
            self.did_header_row_start_with_separator = (
                header_table_row.did_start_with_separator
            )
            self.separator_table_row = separator_table_row
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_header,
            extra_data=None,
            position_marker=position_marker,
            extracted_whitespace="",
            requires_end_token=True,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_header

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownHeaderToken,
            TableMarkdownHeaderToken.__rehydrate_table_header_start,
            TableMarkdownHeaderToken.__rehydrate_table_header_end,
        )

    @staticmethod
    def __rehydrate_table_header_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        context.block_stack.append(current_token)

        header_start_text = cast(
            TableMarkdownHeaderToken, current_token
        ).header_row_leading_whitespace
        return header_start_text + (
            "|"
            if cast(
                TableMarkdownHeaderToken, current_token
            ).did_header_row_start_with_separator
            else ""
        )

    @staticmethod
    def __rehydrate_table_header_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        del context.block_stack[-1]
        start_token = cast(
            TableMarkdownHeaderToken,
            cast(EndMarkdownToken, current_token).start_markdown_token,
        )
        header_end_text = start_token.separator_table_row.extracted_whitespace
        header_end_text += (
            "|" if start_token.separator_table_row.did_start_with_separator else ""
        )
        for i in start_token.separator_table_row.columns:
            header_end_text += i.leading_whitespace + i.text + i.trailing_whitespace
        header_end_text += start_token.separator_table_row.trailing_whitespace
        return (
            start_token.header_row_trailing_whitespace + "\n" + header_end_text + "\n"
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownHeaderToken,
            TableMarkdownHeaderToken.__handle_start_table_token,
            TableMarkdownHeaderToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        return "".join(
            [
                output_html,
                "<thead>",
                ParserHelper.newline_character,
                "<tr>",
                ParserHelper.newline_character,
            ]
        )

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        return "".join(
            [
                output_html,
                "</tr>",
                ParserHelper.newline_character,
                "</thead>",
                ParserHelper.newline_character,
            ]
        )


class TableMarkdownHeaderItemToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the table header item markdown token.
    """

    def __init__(
        self,
        leading_whitespace: str,
        column_alignment: Optional[str],
        position_marker: PositionMarker,
    ) -> None:
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_header_item,
            extra_data=column_alignment,
            position_marker=position_marker,
            extracted_whitespace=leading_whitespace,
            requires_end_token=True,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_header_item

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownHeaderItemToken,
            TableMarkdownHeaderItemToken.__rehydrate_table_header_item_start,
            TableMarkdownHeaderItemToken.__rehydrate_table_header_item_end,
        )

    @staticmethod
    def __rehydrate_table_header_item_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return cast(TableMarkdownHeaderItemToken, current_token).extracted_whitespace

    @staticmethod
    def __rehydrate_table_header_item_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return cast(EndMarkdownToken, current_token).extracted_whitespace

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownHeaderItemToken,
            TableMarkdownHeaderItemToken.__handle_start_table_token,
            TableMarkdownHeaderItemToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        if next_token.extra_data:
            text_to_add = f'<th align="{next_token.extra_data}">'
        else:
            text_to_add = "<th>"
        return "".join(
            [
                output_html,
                text_to_add,
            ]
        )

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        return "".join(
            [
                output_html,
                "</th>",
                ParserHelper.newline_character,
            ]
        )


class TableMarkdownBodyToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the table header markdown token.
    """

    def __init__(
        self,
        position_marker: PositionMarker,
    ) -> None:
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_body,
            extra_data=None,
            position_marker=position_marker,
            extracted_whitespace="",
            requires_end_token=True,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_body

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownBodyToken,
            TableMarkdownBodyToken.__rehydrate_table_body_token_start,
            TableMarkdownBodyToken.__rehydrate_table_body_token_end,
        )

    @staticmethod
    def __rehydrate_table_body_token_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return ""

    @staticmethod
    def __rehydrate_table_body_token_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return ""

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownBodyToken,
            TableMarkdownBodyToken.__handle_start_table_token,
            TableMarkdownBodyToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        return "".join(
            [
                output_html,
                "<tbody>",
                ParserHelper.newline_character,
            ]
        )

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        return "".join([output_html, "</tbody>", ParserHelper.newline_character])


class TableMarkdownRowToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the table header markdown token.
    """

    def __init__(
        self,
        leading_whitespace: str,
        trailing_whitespace: str,
        did_start_with_separator: bool,
        delta: int,
        position_marker: PositionMarker,
    ) -> None:
        self.leading_whitespace = leading_whitespace
        self.trailing_whitespace = trailing_whitespace
        self.did_start_with_separator = did_start_with_separator
        self.delta = delta
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_row,
            extra_data=None,
            position_marker=position_marker,
            extracted_whitespace="",
            requires_end_token=True,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_row

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownRowToken,
            TableMarkdownRowToken.__rehydrate_table_row_start,
            TableMarkdownRowToken.__rehydrate_table_row_end,
        )

    @staticmethod
    def __rehydrate_table_row_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        context.block_stack.append(current_token)
        return cast(TableMarkdownRowToken, current_token).leading_whitespace + (
            "|"
            if cast(TableMarkdownRowToken, current_token).did_start_with_separator
            else ""
        )

    @staticmethod
    def __rehydrate_table_row_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        del context.block_stack[-1]
        return (
            cast(EndMarkdownToken, current_token).extracted_whitespace
            + cast(
                TableMarkdownRowToken,
                cast(EndMarkdownToken, current_token).start_markdown_token,
            ).trailing_whitespace
            + "\n"
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownRowToken,
            TableMarkdownRowToken.__handle_start_table_token,
            TableMarkdownRowToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        return "".join(
            [
                output_html,
                "<tr>",
                ParserHelper.newline_character,
            ]
        )

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        ff = [output_html]
        ff.extend("<td></td>\n" for _ in range(_.start_markdown_token.delta))
        ff.extend(("</tr>", ParserHelper.newline_character))
        return "".join(ff)


class TableMarkdownRowItemToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the table header markdown token.
    """

    def __init__(
        self,
        leading_whitespace: str,
        column_alignment: Optional[str],
        position_marker: PositionMarker,
    ) -> None:
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_row_item,
            extra_data=column_alignment,
            position_marker=position_marker,
            extracted_whitespace=leading_whitespace,
            requires_end_token=True,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_row_item

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownRowItemToken,
            TableMarkdownRowItemToken.__rehydrate_table_row_item_start,
            TableMarkdownRowItemToken.__rehydrate_table_row_item_end,
        )

    @staticmethod
    def __rehydrate_table_row_item_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return cast(TableMarkdownRowItemToken, current_token).extracted_whitespace

    @staticmethod
    def __rehydrate_table_row_item_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return cast(TableMarkdownRowItemToken, current_token).extracted_whitespace

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownRowItemToken,
            TableMarkdownRowItemToken.__handle_start_table_token,
            TableMarkdownRowItemToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        if next_token.extra_data:
            text_to_add = f'<td align="{next_token.extra_data}">'
        else:
            text_to_add = "<td>"
        return "".join(
            [
                output_html,
                text_to_add,
            ]
        )

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        return "".join([output_html, "</td>", ParserHelper.newline_character])



"""
Module to implement a plugin that ensures that top-level lists are surrounded by Blank Lines.
"""

from typing import List, Optional

from pymarkdown.plugin_manager.plugin_details import PluginDetails
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd032(RulePlugin):
    """
    Class to implement a plugin that ensures that top-level lists are surrounded by Blank Lines.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__last_non_end_token: Optional[MarkdownToken] = None
        self.__container_token_stack: List[MarkdownToken] = []
        self.__end_list_end_token: Optional[MarkdownToken] = None

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="blanks-around-lists",
            plugin_id="MD032",
            plugin_enabled_by_default=True,
            plugin_description="Lists should be surrounded by blank lines",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md032.md",
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__last_non_end_token = None
        self.__container_token_stack = []
        self.__end_list_end_token = None

    def __next_token_list_start(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        if (
            self.__last_non_end_token
            and not self.__last_non_end_token.is_blank_line
            and not (
                self.__container_token_stack
                and (
                    self.__container_token_stack[-1].is_list_start
                    or (
                        self.__container_token_stack[-1].is_block_quote_start
                        and self.__container_token_stack[-1].line_number
                        == token.line_number
                    )
                )
            )
        ):
            self.report_next_token_error(context, token)

        self.__container_token_stack.append(token)

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if self.__end_list_end_token:
            if (
                not token.is_blank_line
                and not token.is_new_list_item
                and not token.is_list_end
                and not token.is_block_quote_end
                and not token.is_end_of_stream
            ):
                self.report_next_token_error(context, token, line_number_delta=-1)
            self.__end_list_end_token = None

        if token.is_block_quote_start:
            self.__container_token_stack.append(token)
        elif token.is_block_quote_end:
            del self.__container_token_stack[-1]
        elif token.is_list_start:
            self.__next_token_list_start(context, token)
        elif token.is_list_end:
            assert self.__last_non_end_token is not None
            if not self.__last_non_end_token.is_blank_line:
                self.__end_list_end_token = token
                del self.__container_token_stack[-1]

        if (
            not token.is_end_token
            and not token.is_block_quote_start
            and not token.is_list_start
        ):
            self.__last_non_end_token = token



"""
Module to provide tests related to the basic parts of the scanner.
"""

import logging
import os
import runpy
import sys
from test.markdown_scanner import MarkdownScanner
from test.patches.patch_builtin_open import path_builtin_open_with_exception

from pymarkdown.general.parser_logger import ParserLogger

POGGER = ParserLogger(logging.getLogger(__name__))

if sys.version_info < (3, 10):
    ARGPARSE_X = "optional arguments:"
else:
    ARGPARSE_X = "options:"

if sys.version_info < (3, 13):
    ENABLE_RULES_X = "-e ENABLE_RULES, --enable-rules ENABLE_RULES"
    DISABLE_RULES_X = "-d DISABLE_RULES, --disable-rules DISABLE_RULES"
    CONFIG_FILE_X = "--config CONFIGURATION_FILE, -c CONFIGURATION_FILE"
    SET_CONFIG_X = "--set SET_CONFIGURATION, -s SET_CONFIGURATION"
else:
    ENABLE_RULES_X = "-e, --enable-rules ENABLE_RULES"
    DISABLE_RULES_X = "-d, --disable-rules DISABLE_RULES"
    CONFIG_FILE_X = "--config, -c CONFIGURATION_FILE"
    SET_CONFIG_X = "--set, -s SET_CONFIGURATION"


def test_markdown_with_no_parameters():
    """
    Test to make sure we get the simple information if no parameters are supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = []

    expected_return_code = 2
    expected_output = (
        """usage: main.py [-h] [-e ENABLE_RULES] [-d DISABLE_RULES]
               [--add-plugin ADD_PLUGIN] [--config CONFIGURATION_FILE]
               [--set SET_CONFIGURATION] [--strict-config] [--stack-trace]
               [--continue-on-error]
               [--log-level {CRITICAL,ERROR,WARNING,INFO,DEBUG}]
               [--log-file LOG_FILE] [--return-code-scheme {default,minimal}]
               {extensions,fix,plugins,scan,scan-stdin,version} ...

Lint any found Markdown files.

positional arguments:
  {extensions,fix,plugins,scan,scan-stdin,version}
    extensions          extension commands
    fix                 fix the Markdown files in any specified paths
    plugins             plugin commands
    scan                scan the Markdown files in any specified paths
    scan-stdin          scan the standard input as a Markdown file
    version             version of the application

{ARGPARSE_X}
  -h, --help            show this help message and exit
  {ENABLE_RULES_X}
                        comma separated list of rules to enable
  {DISABLE_RULES_X}
                        comma separated list of rules to disable
  --add-plugin ADD_PLUGIN
                        path to a plugin containing a new rule to apply
  {CONFIG_FILE_X}
                        path to the configuration file to use
  {SET_CONFIG_X}
                        manually set an individual configuration property
  --strict-config       throw an error if configuration is bad, instead of
                        assuming default
  --stack-trace         if an error occurs, print out the stack trace for
                        debug purposes
  --continue-on-error   if a tokenization or plugin error occurs, allow
                        processing to continue
  --log-level {CRITICAL,ERROR,WARNING,INFO,DEBUG}
                        minimum level required to log messages
  --log-file LOG_FILE   destination file for log messages
  --return-code-scheme {default,minimal}
                        scheme to choose for selecting the application return
                        code""".replace(
            "{ARGPARSE_X}", ARGPARSE_X
        )
        .replace("{ENABLE_RULES_X}", ENABLE_RULES_X)
        .replace("{DISABLE_RULES_X}", DISABLE_RULES_X)
        .replace("{CONFIG_FILE_X}", CONFIG_FILE_X)
        .replace("{SET_CONFIG_X}", SET_CONFIG_X)
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_no_parameters_through_module():
    """
    Test to make sure we get the simple information if no parameters are supplied,
    but through the module interface.
    """

    # Arrange
    scanner = MarkdownScanner(use_module=True)
    supplied_arguments = []

    expected_return_code = 2
    expected_output = (
        """usage: __main.py__ [-h] [-e ENABLE_RULES] [-d DISABLE_RULES]
                   [--add-plugin ADD_PLUGIN] [--config CONFIGURATION_FILE]
                   [--set SET_CONFIGURATION] [--strict-config] [--stack-trace]
                   [--continue-on-error]
                   [--log-level {CRITICAL,ERROR,WARNING,INFO,DEBUG}]
                   [--log-file LOG_FILE]
                   [--return-code-scheme {default,minimal}]
                   {extensions,fix,plugins,scan,scan-stdin,version} ...

Lint any found Markdown files.

positional arguments:
  {extensions,fix,plugins,scan,scan-stdin,version}
    extensions          extension commands
    fix                 fix the Markdown files in any specified paths
    plugins             plugin commands
    scan                scan the Markdown files in any specified paths
    scan-stdin          scan the standard input as a Markdown file
    version             version of the application

{ARGPARSE_X}
  -h, --help            show this help message and exit
  {ENABLE_RULES_X}
                        comma separated list of rules to enable
  {DISABLE_RULES_X}
                        comma separated list of rules to disable
  --add-plugin ADD_PLUGIN
                        path to a plugin containing a new rule to apply
  {CONFIG_FILE_X}
                        path to the configuration file to use
  {SET_CONFIG_X}
                        manually set an individual configuration property
  --strict-config       throw an error if configuration is bad, instead of
                        assuming default
  --stack-trace         if an error occurs, print out the stack trace for
                        debug purposes
  --continue-on-error   if a tokenization or plugin error occurs, allow
                        processing to continue
  --log-level {CRITICAL,ERROR,WARNING,INFO,DEBUG}
                        minimum level required to log messages
  --log-file LOG_FILE   destination file for log messages
  --return-code-scheme {default,minimal}
                        scheme to choose for selecting the application return
                        code""".replace(
            "{ARGPARSE_X}", ARGPARSE_X
        )
        .replace("{ENABLE_RULES_X}", ENABLE_RULES_X)
        .replace("{DISABLE_RULES_X}", DISABLE_RULES_X)
        .replace("{CONFIG_FILE_X}", CONFIG_FILE_X)
        .replace("{SET_CONFIG_X}", SET_CONFIG_X)
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_no_parameters_through_main():
    """
    Test to make sure we get the simple information if no parameters are supplied,
    but through the main interface.
    """

    # Arrange
    scanner = MarkdownScanner(use_main=True)
    supplied_arguments = []

    expected_return_code = 2
    expected_output = (
        """usage: main.py [-h] [-e ENABLE_RULES] [-d DISABLE_RULES]
               [--add-plugin ADD_PLUGIN] [--config CONFIGURATION_FILE]
               [--set SET_CONFIGURATION] [--strict-config] [--stack-trace]
               [--continue-on-error]
               [--log-level {CRITICAL,ERROR,WARNING,INFO,DEBUG}]
               [--log-file LOG_FILE] [--return-code-scheme {default,minimal}]
               {extensions,fix,plugins,scan,scan-stdin,version} ...

Lint any found Markdown files.

positional arguments:
  {extensions,fix,plugins,scan,scan-stdin,version}
    extensions          extension commands
    fix                 fix the Markdown files in any specified paths
    plugins             plugin commands
    scan                scan the Markdown files in any specified paths
    scan-stdin          scan the standard input as a Markdown file
    version             version of the application

{ARGPARSE_X}
  -h, --help            show this help message and exit
  {ENABLE_RULES_X}
                        comma separated list of rules to enable
  {DISABLE_RULES_X}
                        comma separated list of rules to disable
  --add-plugin ADD_PLUGIN
                        path to a plugin containing a new rule to apply
  {CONFIG_FILE_X}
                        path to the configuration file to use
  {SET_CONFIG_X}
                        manually set an individual configuration property
  --strict-config       throw an error if configuration is bad, instead of
                        assuming default
  --stack-trace         if an error occurs, print out the stack trace for
                        debug purposes
  --continue-on-error   if a tokenization or plugin error occurs, allow
                        processing to continue
  --log-level {CRITICAL,ERROR,WARNING,INFO,DEBUG}
                        minimum level required to log messages
  --log-file LOG_FILE   destination file for log messages
  --return-code-scheme {default,minimal}
                        scheme to choose for selecting the application return
                        code""".replace(
            "{ARGPARSE_X}", ARGPARSE_X
        )
        .replace("{ENABLE_RULES_X}", ENABLE_RULES_X)
        .replace("{DISABLE_RULES_X}", DISABLE_RULES_X)
        .replace("{CONFIG_FILE_X}", CONFIG_FILE_X)
        .replace("{SET_CONFIG_X}", SET_CONFIG_X)
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_h():
    """
    Test to make sure we get help if '-h' is supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["-h"]

    expected_return_code = 0
    expected_output = (
        """usage: main.py [-h] [-e ENABLE_RULES] [-d DISABLE_RULES]
               [--add-plugin ADD_PLUGIN] [--config CONFIGURATION_FILE]
               [--set SET_CONFIGURATION] [--strict-config] [--stack-trace]
               [--continue-on-error]
               [--log-level {CRITICAL,ERROR,WARNING,INFO,DEBUG}]
               [--log-file LOG_FILE] [--return-code-scheme {default,minimal}]
               {extensions,fix,plugins,scan,scan-stdin,version} ...

Lint any found Markdown files.

positional arguments:
  {extensions,fix,plugins,scan,scan-stdin,version}
    extensions          extension commands
    fix                 fix the Markdown files in any specified paths
    plugins             plugin commands
    scan                scan the Markdown files in any specified paths
    scan-stdin          scan the standard input as a Markdown file
    version             version of the application

{ARGPARSE_X}
  -h, --help            show this help message and exit
  {ENABLE_RULES_X}
                        comma separated list of rules to enable
  {DISABLE_RULES_X}
                        comma separated list of rules to disable
  --add-plugin ADD_PLUGIN
                        path to a plugin containing a new rule to apply
  {CONFIG_FILE_X}
                        path to the configuration file to use
  {SET_CONFIG_X}
                        manually set an individual configuration property
  --strict-config       throw an error if configuration is bad, instead of
                        assuming default
  --stack-trace         if an error occurs, print out the stack trace for
                        debug purposes
  --continue-on-error   if a tokenization or plugin error occurs, allow
                        processing to continue
  --log-level {CRITICAL,ERROR,WARNING,INFO,DEBUG}
                        minimum level required to log messages
  --log-file LOG_FILE   destination file for log messages
  --return-code-scheme {default,minimal}
                        scheme to choose for selecting the application return
                        code""".replace(
            "{ARGPARSE_X}", ARGPARSE_X
        )
        .replace("{ENABLE_RULES_X}", ENABLE_RULES_X)
        .replace("{DISABLE_RULES_X}", DISABLE_RULES_X)
        .replace("{CONFIG_FILE_X}", CONFIG_FILE_X)
        .replace("{SET_CONFIG_X}", SET_CONFIG_X)
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_version():
    """
    Test to make sure we get help if 'version' is supplied.

    This function is shadowed by
    test_markdown_return_code_default_success.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["version"]

    version_path = os.path.join(".", "pymarkdown", "version.py")
    version_meta = runpy.run_path(version_path)
    semantic_version = version_meta["__version__"]

    expected_return_code = 0
    expected_output = """{version}
""".replace(
        "{version}", semantic_version
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_direct_args(caplog):
    """
    Test to make sure we can specify the arguments directly.

    This function is shadowed by
    test_markdown_return_code_default_no_files_to_scan.
    """

    # Arrange
    scanner = MarkdownScanner(use_main=False)
    supplied_arguments = ["--log-level", "DEBUG", "scan", "does-not-exist.md"]

    expected_return_code = 1
    expected_output = ""
    expected_error = """Provided path 'does-not-exist.md' does not exist.


No matching files found."""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, use_direct_arguments=True
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )

    assert "Using direct arguments: [" in caplog.text


def test_markdown_without_direct_args(caplog):
    """
    Test to make sure we can specify the arguments normally.

    This function is shadowed by test_api_scan_for_non_existant_file.
    """

    # Arrange
    scanner = MarkdownScanner(use_main=False)
    supplied_arguments = ["--log-level", "DEBUG", "scan", "does-not-exist.md"]

    expected_return_code = 1
    expected_output = ""
    expected_error = """Provided path 'does-not-exist.md' does not exist.


No matching files found."""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, use_direct_arguments=False
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )

    assert "Using supplied command line arguments." in caplog.text


def test_markdown_with_failure_during_file_scan():
    """
    Test to make sure we get simulate a test scan exception.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "scan",
        source_path,
    ]
    exception_path = os.path.abspath(
        os.path.join("pymarkdown", "resources", "entities.json")
    )

    expected_return_code = 1
    expected_output = ""
    expected_error = """
    
BadTokenizationError encountered while initializing tokenizer:
Named character entity map file '{source_path}' was not loaded (bob).
""".replace(
        "{source_path}", exception_path
    )

    # Act
    with path_builtin_open_with_exception(exception_path, "rt", IOError("bob"), True):
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_x_init():
    """
    Test to make sure we get simulate a test initialization exception if the
    `-x-init` flag is set.

    This function shadows
    test_api_tokenizer_init_exception
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    exception_path = os.path.join("pymarkdown", "resources", "entities.json")
    supplied_arguments = [
        "scan",
        source_path,
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadTokenizationError encountered while initializing tokenizer:\n"
        + f"Named character entity map file '{os.path.abspath(exception_path)}' was not loaded (blah)."
    )

    # Act
    with path_builtin_open_with_exception(
        os.path.abspath(exception_path), "rt", OSError("blah")
    ):
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


# TODO add Markdown parsing of some binary file to cause the tokenizer to throw an exception?


def test_markdown_with_multiple_errors_reported():
    """
    Test to make sure we properly sort errors from files.

    Variation on test_md020_bad_single_paragraph_with_whitespace_at_end
    with no rules disabled.
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test",
        "resources",
        "rules",
        "md020",
        "single_paragraph_with_whitespace_at_end.md",
    )
    expected_return_code = 1
    supplied_arguments = [
        "scan",
        source_path,
    ]

    expected_output = (
        f"{source_path}:1:1: MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + f"{source_path}:1:12: "
        + "MD010: Hard tabs "
        + "[Column: 12] (no-hard-tabs)\n"
        + f"{source_path}:2:2: "
        + "MD021: Multiple spaces are present inside hash characters on Atx Closed Heading. "
        + "(no-multiple-space-closed-atx)\n"
        + f"{source_path}:2:2: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + f"{source_path}:2:2: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
        + f"{source_path}:2:14: "
        + "MD010: Hard tabs "
        + "[Column: 14] (no-hard-tabs)"
    )

    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
https://github.github.com/gfm/#lists
"""

from test.utils import act_and_assert, assert_that_exception_is_raised

import pytest

from pymarkdown.tokens.markdown_token import (
    EndMarkdownToken,
    MarkdownToken,
    MarkdownTokenClass,
)
from pymarkdown.transform_gfm.transform_to_gfm import TransformToGfm


@pytest.mark.gfm
def test_gfm_bad_token():
    """
    Test to ensure that a bad markdown token asserts an error.
    """

    # Arrange
    transformer = TransformToGfm()
    tokens_to_test = [
        MarkdownToken("bad", MarkdownTokenClass.INLINE_BLOCK),
    ]

    expected_output = "Markdown token type <class 'pymarkdown.tokens.markdown_token.MarkdownToken'> not supported."

    # Act & Assert
    assert_that_exception_is_raised(
        AssertionError, expected_output, transformer.transform, tokens_to_test
    )


@pytest.mark.gfm
def test_gfm_bad_end_token():
    """
    Test to ensure that a bad markdown end token asserts an error.
    """

    # Arrange
    transformer = TransformToGfm()
    tokens_to_test = [
        EndMarkdownToken("bad", "", None, "hi", False),
    ]
    expected_output = "Markdown token end type bad not supported."

    # Act & Assert
    assert_that_exception_is_raised(
        AssertionError, expected_output, transformer.transform, tokens_to_test
    )


@pytest.mark.gfm
def test_do_add_end_of_stream_token():
    """
    Test to...

    This function shadows test_atx_headings_033.
    """

    # Arrange
    source_markdown = """####### foo"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):####### foo:]",
        "[end-para:::True]",
        "[end-of-stream(2,0)]",
    ]
    expected_gfm = """<p>####### foo</p>"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, do_add_end_of_stream_token=True
    )
